# Generated by Django 3.2.15 on 2022-09-06 21:25

from django.db import migrations

from pghistory.models import Context


def install_pgh_attach_context_func(apps, schema_editor):
    Context.install_pgh_attach_context_func(using=schema_editor.connection.alias)


class Migration(migrations.Migration):
    dependencies = [
        ("pghistory", "0003_auto_20201023_1636"),
    ]

    operations = [
        migrations.RunPython(
            install_pgh_attach_context_func, reverse_code=migrations.RunPython.noop
        )
    ]



import collections
import contextlib
import json
import threading
import uuid
from typing import Any, Dict, Tuple, Union

from django.db import connection

from pghistory import config, utils

if utils.psycopg_maj_version == 2:
    import psycopg2.extensions
elif utils.psycopg_maj_version == 3:
    import psycopg.pq
else:
    raise AssertionError


_tracker = threading.local()


Context = collections.namedtuple("Context", ["id", "metadata"])


def _is_concurrent_statement(sql: Union[str, bytes]):
    """
    True if the sql statement is concurrent and cannot be ran in a transaction
    """
    sql = sql.strip().lower() if sql else ""
    sql = sql.decode() if isinstance(sql, bytes) else sql
    return sql.startswith("create") and "concurrently" in sql


def _is_transaction_errored(cursor):
    """
    True if the current transaction is in an errored state
    """
    if utils.psycopg_maj_version == 2:
        return (
            cursor.connection.get_transaction_status()
            == psycopg2.extensions.TRANSACTION_STATUS_INERROR
        )
    elif utils.psycopg_maj_version == 3:
        return cursor.connection.info.transaction_status == psycopg.pq.TransactionStatus.INERROR
    else:
        raise AssertionError


def _can_inject_variable(cursor, sql):
    """True if we can inject a SQL variable into a statement.

    A named cursor automatically prepends
    "NO SCROLL CURSOR WITHOUT HOLD FOR" to the query, which
    causes invalid SQL to be generated. There is no way
    to override this behavior in psycopg, so ignoring triggers
    cannot happen for named cursors. Django only names cursors
    for iterators and other statements that read the database,
    so it seems to be safe to ignore named cursors.

    Concurrent index creation is also incompatible with local variable
    setting. Ignore these cases for now.
    """
    return (
        not getattr(cursor, "name", None)
        and not _is_concurrent_statement(sql)
        and not _is_transaction_errored(cursor)
    )


def _execute_wrapper(execute_result):
    if utils.psycopg_maj_version == 3:
        while execute_result is not None and execute_result.nextset():
            pass
    return execute_result


def _inject_history_context(
    execute, sql: Union[str, bytes], params: Union[Dict[str, Any], Tuple[Any, ...]], many, context
):
    is_bytes = isinstance(sql, bytes)
    sql = sql.decode() if is_bytes else sql
    inject_vars = ""

    if _can_inject_variable(context["cursor"], sql):
        # Metadata is stored as a serialized JSON string with escaped
        # single quotes
        serialized_metadata = json.dumps(_tracker.value.metadata, cls=config.json_encoder())
        context_params = {
            "pghistory__context_id": str(_tracker.value.id),
            "pghistory__context_metadata": serialized_metadata,
        }

        # psycopg does not allow params to be mixed (named and series), so we
        # try to preserve what it was.
        if isinstance(params, dict):
            id_placeholder = "%(pghistory__context_id)s"
            metadata_placeholder = "%(pghistory__context_metadata)s"
            params.update(context_params)

        else:
            id_placeholder = metadata_placeholder = "%s"
            params = (*context_params.values(), *(params or ()))

        inject_vars = (
            f"SELECT set_config('pghistory.context_id', {id_placeholder}, true), "
            f"set_config('pghistory.context_metadata', {metadata_placeholder}, true); "
        )

    sql = inject_vars + sql
    sql = sql.encode() if is_bytes else sql
    return _execute_wrapper(execute(sql, params, many, context))


class context(contextlib.ContextDecorator):
    """
    A context manager that groups changes under the same context and
    adds additional metadata about the event.

    Context is added as variables at the beginning of every SQL statement.
    By default, all variables are localized to the transaction (i.e
    SET LOCAL), meaning they will only persist for the statement/transaction
    and not across the session.

    Once any code has entered [pghistory.context][], all subsequent
    entrances of [pghistory.context][] will be grouped under the same
    context until the top-most parent exits.

    To add context only if a parent has already entered [pghistory.context][],
    one can call [pghistory.context][] as a function without entering it.
    The metadata set in the function call will be part of the context if
    [pghistory.context][] has previously been entered. Otherwise it will
    be ignored.

    Attributes:
        **metadata: Metadata that should be attached to the tracking
            context

    Example:
        Here we track a "key" with a value of "value":

            with pghistory.context(key='value'):
                # Do things..
                # All tracked events will have the same `pgh_context`
                # foreign key, and the context object will include
                # {'key': 'value'} in its metadata.
                # Nesting the tracker adds additional metadata to the current
                # context

            # Add metadata if a parent piece of code has already entered
            # pghistory.context
            pghistory.context(key='value')

    Note:
        Context tracking is compatible for most scenarios, but it currently
        does not work for named cursors. Django uses named cursors for
        the .iterator() operator, which has no effect on history tracking.
        However, there may be other usages of named cursors in Django where
        history context is ignored.
    """

    def __init__(self, **metadata: Any):
        self.metadata = metadata
        self._pre_execute_hook = None

        if hasattr(_tracker, "value"):
            _tracker.value.metadata.update(**self.metadata)

    def __enter__(self):
        if not hasattr(_tracker, "value"):
            self._pre_execute_hook = connection.execute_wrapper(_inject_history_context)
            self._pre_execute_hook.__enter__()
            _tracker.value = Context(id=uuid.uuid4(), metadata=self.metadata)

        return _tracker.value

    def __exit__(self, *exc):
        if self._pre_execute_hook:
            delattr(_tracker, "value")
            self._pre_execute_hook.__exit__(*exc)



"""Core way to access configuration"""

import copy
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union

from django.apps import apps
from django.conf import settings
from django.db import models
from django.utils.module_loading import import_string

from pghistory import constants

if TYPE_CHECKING:
    from django.core.serializers.json import DjangoJSONEncoder
    from django.db.models import Model, QuerySet

    from pghistory.admin import EventsAdmin
    from pghistory.core import Tracker


def default_trackers() -> Union[Tuple["Tracker", ...], None]:
    """
    The default event trackers.

    Returns:
        Default pghistory trackers
    """
    default_trackers = getattr(settings, "PGHISTORY_DEFAULT_TRACKERS", None)

    # Copy the default trackers, otherwise we end up with the same instances used across
    # all models which causes issues
    if default_trackers:
        default_trackers = copy.deepcopy(default_trackers)

    return default_trackers


def append_only() -> bool:
    """If event models should be append-only by default.

    Protection triggers are installed for update and delete operations if `True`.

    Returns:
        `True` if event models should be append-only by default.
    """
    return getattr(settings, "PGHISTORY_APPEND_ONLY", False)


def middleware_methods() -> Tuple[str, ...]:
    """
    Methods tracked by the pghistory middleware.

    Returns:
        The HTTP methods
    """
    return getattr(
        settings, "PGHISTORY_MIDDLEWARE_METHODS", ("GET", "POST", "PUT", "PATCH", "DELETE")
    )


def install_context_func_on_migrate() -> bool:
    """True if the pghistory context tracking function is installed after migration.

    Defaults to `False`.

    Returns:
        `True` if the _pgh_attach_context() Postgres func is installed after migration.
    """
    return getattr(settings, "PGHISTORY_INSTALL_CONTEXT_FUNC_ON_MIGRATE", False)


def json_encoder() -> "DjangoJSONEncoder":
    """The JSON encoder when tracking context

    Returns:
        The JSON encoder class
    """
    encoder = getattr(
        settings, "PGHISTORY_JSON_ENCODER", "django.core.serializers.json.DjangoJSONEncoder"
    )

    if isinstance(encoder, str):  # pragma: no branch
        encoder = import_string(encoder)

    return encoder


def base_model() -> Type["Model"]:
    """The base model for event models.

    Returns:
        The model class
    """
    event_model = import_string("pghistory.models.Event")
    base_model = getattr(settings, "PGHISTORY_BASE_MODEL", event_model)

    if isinstance(base_model, str):  # pragma: no cover
        base_model = import_string(base_model)

    assert issubclass(base_model, event_model)
    return base_model


def field() -> "Field":
    """The default configuration for all fields in event models.

    Returns:
        The pghistory `Field` class
    """
    field = getattr(settings, "PGHISTORY_FIELD", Field())
    assert isinstance(field, Field)
    return field


def related_field() -> "RelatedField":
    """The default configuration for related fields in event models.

    Returns:
        The pghistory `RelatedField` class
    """
    related_field = getattr(settings, "PGHISTORY_RELATED_FIELD", RelatedField())
    assert isinstance(related_field, RelatedField)
    return related_field


def foreign_key_field() -> "ForeignKey":
    """The default configuration for foreign keys in event models.

    Returns:
        The pghistory `ForeignKey` class
    """
    foreign_key_field = getattr(settings, "PGHISTORY_FOREIGN_KEY_FIELD", ForeignKey())
    assert isinstance(foreign_key_field, ForeignKey)
    return foreign_key_field


def context_field() -> Union["ContextForeignKey", "ContextJSONField"]:
    """The default field config to use for context in event models.

    Returns:
        The pghistory field class
    """

    # Note: We will be changing the default context field to have on_delete=PROTECT
    # in version 3.
    context_field = getattr(settings, "PGHISTORY_CONTEXT_FIELD", ContextForeignKey())
    assert isinstance(context_field, (ContextForeignKey, ContextJSONField))
    return context_field


def context_id_field() -> "ContextUUIDField":
    """
    The default field config to use for context ID in event models when context is denormalized.

    Returns:
        The pghistory context UUID field
    """

    context_id_field = getattr(settings, "PGHISTORY_CONTEXT_ID_FIELD", ContextUUIDField())
    assert isinstance(context_id_field, ContextUUIDField)
    return context_id_field


def obj_field() -> "ObjForeignKey":
    """The default field config to use for object references in event models.

    Returns:
        The pghistory object field
    """

    obj_field = getattr(settings, "PGHISTORY_OBJ_FIELD", ObjForeignKey())
    assert isinstance(obj_field, ObjForeignKey)
    return obj_field


def exclude_field_kwargs() -> Dict["Field", List[str]]:
    """
    Provide a mapping of field classes to a list of keyword args to ignore
    when instantiating the field on the event model.

    For example, a field may not allow `unique` as a keyword argument.
    If so, set `settings.PGHISTORY_EXCLUDE_FIELD_KWARGS = {"field.FieldClass": ["unique"]}`

    Returns:
        The mapping of field classes to kwargs
    """
    exclude_field_kwargs = getattr(settings, "PGHISTORY_EXCLUDE_FIELD_KWARGS", {})
    assert isinstance(exclude_field_kwargs, dict)
    for val in exclude_field_kwargs.values():
        assert isinstance(val, (list, tuple))

    # Import strings
    exclude_field_kwargs = {
        import_string(key) if isinstance(key, str) else key: value
        for key, value in exclude_field_kwargs.items()
    }

    return exclude_field_kwargs


def created_at_function() -> str:
    """The default PostgreSQL function used to populate the "pgh_created_at" field.

    Use `clock_timestamp()` to set the actual current time and not the time of the transaction.

    Returns:
        PostgreSQL function to be used as a default value for the "pgh_created_at" field.
    """
    return getattr(settings, "PGHISTORY_CREATED_AT_FUNCTION", "NOW()")


def admin_ordering() -> List[str]:
    """The default ordering for the events admin.

    Returns:
        The fields over which to order by
    """
    ordering = getattr(settings, "PGHISTORY_ADMIN_ORDERING", "-pgh_created_at") or []

    if not isinstance(ordering, (list, tuple)):
        ordering = [ordering]

    return ordering


def admin_model() -> Type["Model"]:
    """The default list display for the events admin.

    Returns:
        The model class
    """
    return apps.get_model(getattr(settings, "PGHISTORY_ADMIN_MODEL", "pghistory.Events"))


def admin_queryset() -> "QuerySet":
    """The default queryset for the events admin.

    Returns:
        The queryset
    """
    return getattr(
        settings, "PGHISTORY_ADMIN_QUERYSET", admin_model().objects.order_by(*admin_ordering())
    )


def admin_class() -> Type["EventsAdmin"]:
    """The admin class to use for the events admin.

    Returns:
        The pghistory.admin.EventsAdmin class
    """

    events_admin = import_string("pghistory.admin.EventsAdmin")
    admin_class = getattr(settings, "PGHISTORY_ADMIN_CLASS", events_admin)

    if isinstance(admin_class, str):  # pragma: no cover
        admin_class = import_string(admin_class)

    assert issubclass(admin_class, events_admin)

    return admin_class


def admin_all_events() -> bool:
    """True if all events should be shown in the admin when there are no filters.

    Returns:
        The bool setting
    """
    return getattr(settings, "PGHISTORY_ADMIN_ALL_EVENTS", True)


def admin_list_display() -> List[str]:
    """The default list display for the events admin.

    Returns:
        The field display order list
    """
    defaults = ["pgh_created_at", "pgh_obj_model", "pgh_obj_id", "pgh_diff"]

    if admin_queryset().model._meta.label == "pghistory.MiddlewareEvents":
        defaults.extend(["user", "url"])

    return getattr(settings, "PGHISTORY_ADMIN_LIST_DISPLAY", defaults)


def _get_kwargs(vals):
    return {
        key: val
        for key, val in vals.items()
        if key not in ("self", "kwargs", "__class__") and val is not constants.UNSET
    }


class Field:
    """Configuration for fields.

    The default values for the attributes ensure that
    event models don't have unnecessary uniqueness constraints
    carried over from the tracked model.

    Attributes:
        primary_key (bool, default=False): True if a primary key
        unique (bool, default=False): True if unique
        blank (bool): True if blank
        null (bool): True if null
        db_index (bool): True if indexed
        editable (bool): True if editable
        unique_for_date (bool): True if unique for date
        unique_for_month (bool): True if unique for the month
        unique_for_year (bool): True if unique for the year
    """

    def __init__(
        self,
        *,
        primary_key: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        unique: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        blank: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        null: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        db_index: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        editable: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        unique_for_date: Union[bool, None, constants.Unset, constants.Default] = constants.UNSET,
        unique_for_month: Union[bool, None, constants.Unset, constants.Default] = constants.UNSET,
        unique_for_year: Union[bool, None, constants.Unset, constants.Default] = constants.UNSET,
    ):
        self._kwargs = _get_kwargs(locals())
        self._finalized = False

    @property
    def kwargs(self):
        return {
            key: val
            for key, val in {**self.get_default_kwargs(), **self._kwargs}.items()
            if val is not constants.DEFAULT
        }

    def get_default_kwargs(self):
        return {
            **Field(
                primary_key=False,
                unique=False,
                db_index=False,
                unique_for_date=None,
                unique_for_month=None,
                unique_for_year=None,
            )._kwargs,
            **field()._kwargs,
        }


class RelatedField(Field):
    """Configuration for related fields.

    By default, related names are stripped to avoid
    unnecessary clashes.

    Note that all arguments from `Field` can also be supplied.

    Attributes:
        related_name (str, default="+"): The related_name to use
        related_query_name (str, default="+"): The related_query_name
            to use
    """

    def __init__(
        self,
        *,
        related_name: Union[str, constants.Unset, constants.Default] = constants.UNSET,
        related_query_name: Union[str, constants.Unset, constants.Default] = constants.UNSET,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._kwargs.update(_get_kwargs(locals()))

    def get_default_kwargs(self):
        return {
            **super().get_default_kwargs(),
            **RelatedField(related_name="+", related_query_name="+")._kwargs,
            **related_field()._kwargs,
        }


class ForeignKey(RelatedField):
    """Configuration for foreign keys.

    Arguments for `RelatedField` and `Field` can also be supplied.
    Note that db_index is overridden to `True` for all foreign keys

    Attributes:
        on_delete (default=models.DO_NOTHING): Django's on_delete property
        db_constraint (bool, default=False): True to use a datbase constraint
            for the foreign key
    """

    def __init__(
        self,
        *,
        on_delete: Any = constants.UNSET,
        db_constraint: Union[bool, constants.Unset, constants.Default] = constants.UNSET,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self._kwargs.update(_get_kwargs(locals()))

    def get_default_kwargs(self):
        return {
            **super().get_default_kwargs(),
            **ForeignKey(on_delete=models.DO_NOTHING, db_index=True, db_constraint=False)._kwargs,
            **foreign_key_field()._kwargs,
        }


class ContextForeignKey(ForeignKey):
    """Configuration for the `pgh_context` field when a foreign key is used.

    Overrides null to `True`.

    Attributes:
        null (bool, default=True): True if nullable context is allowed
        related_query_name (str): The related_query_name to use
    """

    def __init__(
        self,
        *,
        null: bool = True,
        related_query_name: Union[str, constants.Unset, constants.Default] = constants.DEFAULT,
        **kwargs: Any,
    ):
        super().__init__(null=null, related_query_name=related_query_name, **kwargs)
        self._kwargs.update(_get_kwargs(locals()))


class ContextJSONField(Field):
    """Configuration for the `pgh_context` field when denormalized context is used.

    Attributes:
        null (bool, default=True): True if nullable context is allowed
    """

    def __init__(self, *, null: bool = True, **kwargs: Any):
        super().__init__(null=null, **kwargs)
        self._kwargs.update(_get_kwargs(locals()))


class ContextUUIDField(Field):
    """Configuration for the `pgh_context_id` field when denormalized context is used.

    Attributes:
        null (bool, default=True): True if nullable context is allowed
    """

    def __init__(self, *, null: bool = True, **kwargs: Any):
        super().__init__(null=null, **kwargs)
        self._kwargs.update(_get_kwargs(locals()))


class ObjForeignKey(ForeignKey):
    """Configuration for the `pgh_obj` field

    Attributes:
        related_name (str): The related_name to use
        related_query_name (str): The related_query_name to use
    """

    def __init__(
        self,
        *,
        related_name: Union[str, constants.Unset, constants.Default] = constants.DEFAULT,
        related_query_name: Union[str, constants.Unset, constants.Default] = constants.DEFAULT,
        **kwargs,
    ):
        # Note: We will be changing the default object field to nullable with on_delete=SET_NULL
        # in version 3. related_name will also default to "+"
        super().__init__(
            related_name=related_name, related_query_name=related_query_name, **kwargs
        )
        self._kwargs.update(_get_kwargs(locals()))



from django.conf import settings
from django.core import checks


@checks.register(checks.Tags.compatibility)
def check_pgtrigger_installed(app_configs, **kwargs):
    errors = []

    if "pgtrigger" not in settings.INSTALLED_APPS:
        errors.append(
            checks.Error(
                'Add "pgtrigger" to settings.INSTALLED_APPS to use django-pghistory.',
                id="pghistory.E001",
            )
        )

    return errors



import warnings

import django
from django.apps import apps
from django.contrib import admin
from django.contrib.admin.views.main import ChangeList
from django.utils.encoding import force_str

from pghistory import config, core


def _get_model(model):
    if model:
        try:
            return apps.get_model(model)
        except Exception:
            pass


def _get_obj(model_pk):
    """Gets an object from a model:pk string"""
    if model_pk and ":" in model_pk:
        model, pk = model_pk.split(":", 1)
        model = _get_model(model)
        if model:
            return model(pk=pk)


class MethodFilter(admin.SimpleListFilter):
    title = "method"
    parameter_name = "method"

    def lookups(self, request, model_admin):
        return (("tracks", "tracks"), ("references", "references"))

    def queryset(self, request, queryset):
        obj = request.GET.get("obj")
        obj = _get_obj(obj)
        if obj and self.value() == "tracks":
            queryset = queryset.tracks(obj)
        elif obj and self.value() == "references":  # pragma: no branch
            queryset = queryset.references(obj)

        return queryset

    def choices(self, changelist):
        for lookup, title in self.lookup_choices:
            yield {
                "selected": self.value() == force_str(lookup),
                "query_string": changelist.get_query_string({self.parameter_name: lookup}, []),
                "display": title,
            }


def _filter_event_models(request):
    obj = _get_obj(request.GET.get("obj"))
    method = request.GET.get("method")
    model = _get_model(request.GET.get("model"))

    if obj and method == "tracks":
        return core.event_models(tracks_model=obj.__class__)
    elif obj and method == "references":
        return core.event_models(references_model=obj.__class__)
    elif model:
        return core.event_models(tracks_model=model, include_missing_pgh_obj=True)
    else:
        return core.event_models()


class LabelFilter(admin.SimpleListFilter):
    title = "label"
    parameter_name = "label"

    def lookups(self, request, model_admin):
        from pghistory.models import Event  # noqa

        if issubclass(model_admin.model, Event):
            event_models = [model_admin.model]
        else:
            event_models = _filter_event_models(request)

        labels = {
            tracker.label
            for event_model in event_models
            for tracker in event_model.pgh_trackers or []
        }
        return sorted([(label, label) for label in labels])

    def queryset(self, request, queryset):
        if self.value():
            queryset = queryset.filter(pgh_label=self.value())

        return queryset


class EventModelFilter(admin.SimpleListFilter):
    title = "event model"
    parameter_name = "event_model"

    def lookups(self, request, model_admin):
        event_models = _filter_event_models(request)
        return sorted(
            (event_model._meta.label_lower, event_model._meta.label)
            for event_model in event_models
        )

    def queryset(self, request, queryset):
        model = _get_model(request.GET.get("model"))
        if self.value():
            queryset = queryset.across(self.value())
        elif model:
            queryset = queryset.across(*_filter_event_models(request))

        return queryset


class DynamicFilter(admin.SimpleListFilter):
    title = ""

    def queryset(self, request, queryset):
        return queryset

    def lookups(self, request, model_admin):
        if self.parameter_name in request.GET:
            return [(request.GET[self.parameter_name], request.GET[self.parameter_name])]
        else:  # pragma: no cover
            return []


class ObjFilter(DynamicFilter):
    title = "object"
    parameter_name = "obj"

    def queryset(self, request, queryset):
        if (
            "method" not in request.GET
            and hasattr(queryset.model, "pgh_obj")
            and ":" in self.value()
        ):
            queryset = queryset.filter(pgh_obj=self.value().split(":", 1)[1])

        return queryset


class ModelFilter(DynamicFilter):
    title = "model"
    parameter_name = "model"


class BackFilter(DynamicFilter):
    parameter_name = "back"
    template = "pghistory_admin/hidden_filter.html"


class BaseEventAdmin(admin.ModelAdmin):
    change_list_template = "pghistory_admin/events_change_list.html"

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def changelist_view(self, request, extra_context=None):
        extra_context = {"title": self.model._meta.verbose_name_plural.capitalize()}
        obj = _get_obj(request.GET.get("obj"))
        method = request.GET.get("method")
        model = _get_model(request.GET.get("model"))
        if obj:
            obj.refresh_from_db()
            if method:
                extra_context["subtitle"] = f"{method.title()} {obj}"
            else:
                extra_context["subtitle"] = f"{obj}"
        elif model:
            extra_context["subtitle"] = f"{model._meta.label}"

        if "back" in request.GET:
            extra_context["pgh_back"] = request.GET["back"]

        return super().changelist_view(request, extra_context=extra_context)


class EventModelAdmin(BaseEventAdmin):
    """The base admin for event models"""

    list_filter = [LabelFilter, ObjFilter, BackFilter]


class EventsChangeList(ChangeList):
    def get_queryset(self, request):
        # Note: Call get_queryset first so that has_active_filters is accurate
        qset = super().get_queryset(request)

        if not config.admin_all_events():
            if django.VERSION < (3, 1):  # pragma: no cover
                warnings.warn(
                    "PGHISTORY_ADMIN_ALL_EVENTS only works for Django 3.1 and above", stacklevel=2
                )
            elif not self.has_active_filters:  # pragma: no branch
                return self.root_queryset.model.no_objects.all()

        return qset


class EventsAdmin(BaseEventAdmin):
    """
    The admin for showing events across all event models
    """

    def get_changelist(self, request, **kwargs):
        return EventsChangeList

    def get_list_display(self, request):
        return config.admin_list_display()

    def get_list_filter(self, request):
        filters = [LabelFilter, EventModelFilter]
        if "obj" in request.GET:
            filters.extend([MethodFilter, ObjFilter, BackFilter])
        elif "model" in request.GET:
            filters.extend([ModelFilter, BackFilter])

        return filters

    def get_queryset(self, request):
        return config.admin_queryset()



import uuid
import warnings

import django
from django.apps import apps
from django.conf import settings
from django.db import DEFAULT_DB_ALIAS, connections, models
from django.db.models.functions import Cast
from django.db.models.sql import Query
from django.db.models.sql.compiler import SQLCompiler

from pghistory import core, utils

# This class is to preserve backwards compatibility with migrations
PGHistoryJSONField = utils.JSONField


class Context(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    metadata = utils.JSONField(default=dict)

    @classmethod
    def install_pgh_attach_context_func(cls, using: str = DEFAULT_DB_ALIAS) -> None:
        """
        Installs a custom store procedure for upserting context
        for historical events. The upsert is aware of when tracking is
        enabled in the app (i.e. using pghistory.context())

        This stored procedure is automatically installed in pghistory migration 0004.
        """
        connection = connections[using]
        if not connection.vendor.startswith("postgres"):  # pragma: no cover
            return

        with connection.cursor() as cursor:
            cursor.execute(
                f"""
                CREATE OR REPLACE FUNCTION _pgh_attach_context()
                RETURNS {cls._meta.db_table}.id%TYPE AS $$
                    DECLARE
                        _pgh_context_id UUID;
                        _pgh_context_metadata JSONB;
                    BEGIN
                        BEGIN
                            SELECT INTO _pgh_context_id
                                CURRENT_SETTING('pghistory.context_id');
                            SELECT INTO _pgh_context_metadata
                                CURRENT_SETTING('pghistory.context_metadata');
                            EXCEPTION WHEN OTHERS THEN
                        END;
                        IF _pgh_context_id IS NOT NULL AND _pgh_context_metadata IS NOT NULL THEN
                            INSERT INTO {cls._meta.db_table} (id, metadata, created_at, updated_at)
                                VALUES (_pgh_context_id, _pgh_context_metadata, NOW(), NOW())
                                ON CONFLICT (id) DO UPDATE
                                    SET metadata = EXCLUDED.metadata,
                                        updated_at = EXCLUDED.updated_at;
                            RETURN _pgh_context_id;
                        ELSE
                            RETURN NULL;
                        END IF;
                    END;
                $$ LANGUAGE plpgsql;
                """
            )


class EventQueryCompiler(SQLCompiler):
    def _get_cte(self):
        """
        Returns a CTE that selects from proxied fields
        """
        event_model = self.query.model

        annotations = {
            f"_{field.column}": Cast(field.pgh_proxy, output_field=field)
            for field in self.proxy_fields
        }
        values = [
            field.name for field in event_model._meta.fields if not hasattr(field, "pgh_proxy")
        ] + [f"_{field.column}" for field in self.proxy_fields]

        qset = models.QuerySet(event_model).annotate(**annotations).values(*values)

        sql, params = qset.query.as_sql(self, self.connection)
        with self.connection.cursor() as cursor:
            sql = cursor.mogrify(sql, params)
            if isinstance(sql, bytes):  # psycopg 2/3 return different types
                sql = sql.decode("utf-8")

        for field in self.proxy_fields:
            sql = sql.replace(f"_{field.column}", field.column)

        sql = sql.replace("->", "->>")
        return "WITH pgh_event_cte AS (\n" + sql + "\n)\n"

    @property
    def proxy_fields(self):
        return [f for f in self.query.model._meta.fields if hasattr(f, "pgh_proxy")]

    def as_sql(self, *args, **kwargs):
        """
        If there are proxied fields on the event model, return a select from a CTE.
        Otherwise don't do anything special
        """
        sql, params = super().as_sql(*args, **kwargs)

        if any(self.proxy_fields):
            if django.VERSION < (3, 2):  # pragma: no cover
                raise RuntimeError("Must use Django 3.2 or above to proxy fields on event models")

            cte = self._get_cte()
            sql = cte + sql.replace(f'"{self.query.model._meta.db_table}"', '"pgh_event_cte"')

        return sql, params


class EventQuery(Query):
    """A query over an event CTE when proxy fields are used"""

    def get_compiler(self, *args, **kwargs):
        """
        Overrides the Query method get_compiler in order to return
        an EventQueryCompiler.
        """
        compiler = super().get_compiler(*args, **kwargs)
        compiler.__class__ = EventQueryCompiler
        return compiler


class EventQuerySet(models.QuerySet):
    """QuerySet with support for proxy fields"""

    def __init__(self, model=None, query=None, using=None, hints=None):
        if query is None:
            query = EventQuery(model)

        super().__init__(model, query, using, hints)


class PghEventModel:
    "A descriptor for accessing the pgh_event_model field on a tracked model"

    def __get__(self, instance, owner):
        event_models = set(owner.pgh_event_models.values())
        if len(event_models) == 1:
            return list(event_models)[0]
        else:
            raise ValueError(
                f"{owner.__name__} has more than one event model."
                " Use the pgh_event_models dictionary to retrieve the event"
                " model by label."
            )


class Event(models.Model):
    """
    An abstract model for base elements of a event
    """

    pgh_id = models.AutoField(primary_key=True)
    pgh_created_at = models.DateTimeField(auto_now_add=True)
    pgh_label = models.TextField(help_text="The event label.")
    pgh_trackers = None
    pgh_tracked_model = None

    objects = EventQuerySet.as_manager()

    class Meta:
        abstract = True

    @property
    def can_revert(self):
        """True if the event model can revert the tracked model"""
        model_fields = {f.name for f in self.pgh_tracked_model._meta.fields}
        tracked_fields = {f.name for f in self._meta.fields}
        return model_fields.issubset(tracked_fields)

    def revert(self, using=DEFAULT_DB_ALIAS):
        """
        Reverts the tracked model based on the event fields.

        Raises a RuntimeError if the event model doesn't track all fields
        """
        if not self.can_revert:
            raise RuntimeError(
                f'Event model "{self.__class__.__name__}" cannot revert'
                f' "{self.pgh_tracked_model.__class__.__name__}" because it'
                " doesn't track every field."
            )

        qset = models.QuerySet(model=self.pgh_tracked_model, using=using)

        pk = getattr(self, self.pgh_tracked_model._meta.pk.name)
        return qset.update_or_create(
            pk=pk,
            defaults={
                field.name: getattr(self, field.name)
                for field in self.pgh_tracked_model._meta.fields
                if field != self.pgh_tracked_model._meta.pk
            },
        )[0]

    @classmethod
    def pghistory_setup(cls):
        """
        Called when the class is prepared (see apps.py)
        to finalize setup of the model and register triggers
        """
        if (
            not cls._meta.abstract and cls._meta.managed and not cls._meta.proxy
        ):  # pragma: no branch
            for tracker in cls.pgh_trackers or []:
                tracker.pghistory_setup(cls)

            # Set up the event model utility properties. Don't overwrite any
            # existing attributes
            if not hasattr(cls.pgh_tracked_model, "pgh_event_models"):
                cls.pgh_tracked_model.pgh_event_models = {}

            # Use dir here, otherwise the property is evaluated
            if "pgh_event_model" not in dir(cls.pgh_tracked_model):
                cls.pgh_tracked_model.pgh_event_model = PghEventModel()

            if isinstance(cls.pgh_tracked_model.pgh_event_models, dict):  # pragma: no branch
                cls.pgh_tracked_model.pgh_event_models.update(
                    {tracker.label: cls for tracker in cls.pgh_trackers or []}
                )

    @classmethod
    def check(cls, **kwargs):
        """
        Allow proxy models to inherit this model and define their own fields
        that are dynamically pulled from context
        """
        errors = super().check(**kwargs)

        # If all local fields are proxied, ignored error E017 and allow the fields to be declared
        if any(error for error in errors if error.id == "models.E017") and all(
            hasattr(field, "pgh_proxy") for field in cls._meta.local_fields
        ):
            return [error for error in errors if error.id != "models.E017"]
        else:
            return errors


class EventsQueryCompiler(SQLCompiler):
    def _get_empty_select(self):
        """
        When targetting a model that has no event tables, there are
        no valid tables from which a CTE can be generated.

        This method generates a CTE that returns an empty table in the
        schema of the Events table. Note that it's impossible to
        create an empty CTE, so we select NULL VALUES and LIMIT to 0.
        """
        col_name_clause = ", ".join([field.column for field in self.query.model._meta.fields])
        col_select_clause = ",\n".join(
            [
                f"_pgh_obj_event.{field.column}::"
                f"{field.rel_db_type(self.connection)} AS {field.attname}"
                for field in self.query.model._meta.fields
            ]
        )
        values_list = ["(NULL)" for field in self.query.model._meta.fields]
        return f"""
            SELECT
              {col_select_clause}
            FROM (
              VALUES ({', '.join(values_list)}) LIMIT 0
            ) AS _pgh_obj_event({col_name_clause})
            WHERE pgh_model IS NOT NULL
        """

    def _validate(self):
        if (
            isinstance(self.references, (list, tuple))
            and len({r.__class__ for r in self.references}) > 1
        ):
            raise ValueError("The objects passed to references() are not of the same type.")
        elif (
            isinstance(self.tracks, (list, tuple)) and len({o.__class__ for o in self.tracks}) > 1
        ):
            raise ValueError("The objects passed to tracks() are not of the same type.")
        elif self.references_model and self.tracks_model:
            raise ValueError("Cannot use both tracks() and references().")

    @property
    def references_model(self):
        if isinstance(self.references, models.QuerySet):
            return self.references.model._meta.concrete_model
        elif isinstance(self.references, (list, tuple)) and self.references:
            return self.references[0].__class__._meta.concrete_model

    @property
    def references(self):
        return self.query.references

    @property
    def tracks_model(self):
        if isinstance(self.tracks, models.QuerySet):
            return self.tracks.model._meta.concrete_model
        elif isinstance(self.tracks, (list, tuple)) and self.tracks:
            return self.tracks[0].__class__._meta.concrete_model

    @property
    def tracks(self):
        return self.query.tracks

    @property
    def across(self):
        return core.event_models(
            models=self.query.across,
            references_model=self.references_model,
            tracks_model=self.tracks_model,
        )

    def _get_context_clauses(self, event_model):
        """
        Get the clauses for obtaining context based on the event model

        We have the following cases to handle:
        1. No pgh_context
        2. A pgh_context foreign key is used
        3. A pgh_context JSON is used with pgh_context_id
        4. A pgh_context JSON is used without pgh_context_id
        """
        proxy_fields = []
        for field in self.query.model._meta.fields:
            if hasattr(field, "pgh_proxy"):
                if not field.pgh_proxy.startswith("pgh_context__"):  # pragma: no cover
                    raise RuntimeError(
                        "Proxy fields on Events models can only proxy the pgh_context field."
                        " E.g. pgh_context__url"
                    )

                proxy_fields.append((field, field.pgh_proxy.split("__", 1)[1]))
            elif not field.attname.startswith("pgh_"):
                warnings.warn(
                    f"django-pghistory extra field '{field}' in event model"
                    f" '{self.query.model._meta.label}' declared. Use"
                    " 'pghistory.ProxyField' to define a proxy fields instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                proxy_fields.append((field, field.name))

        context_join_clause = ""
        final_context_columns_clause = "".join(
            [f"_pgh_obj_event.{field.column},\n" for field, _ in proxy_fields]
        )

        if not hasattr(event_model, "pgh_context"):
            context_id_column_clause = "NULL::UUID AS pgh_context_id"
            context_column_clause = "NULL::JSONB AS pgh_context"

            # If the aggregate event model has any proxy fields,
            # make them null since there is no context on this event
            annotated_context_columns_clause = "".join(
                [
                    f"NULL::{field.rel_db_type(self.connection)} AS {field.column},\n"
                    for field, _ in proxy_fields
                ]
            )
        elif isinstance(event_model._meta.get_field("pgh_context"), models.ForeignKey):
            context_id_column_clause = "pgh_context_id"
            context_column_clause = "_pgh_context.metadata AS pgh_context"

            # If the aggregate event model has any proxy fields,
            # pull these directly from the context metadata
            annotated_context_columns_clause = "".join(
                [
                    f"(_pgh_context.metadata->>'{attr}')::"
                    f"{field.rel_db_type(self.connection)} AS {field.column},\n"
                    for field, attr in proxy_fields
                ]
            )
            context_join_clause = f"""
                LEFT OUTER JOIN {Context._meta.db_table} _pgh_context
                    ON _pgh_context.id = _event.pgh_context_id
            """
        elif isinstance(event_model._meta.get_field("pgh_context"), utils.DjangoJSONField):
            context_column_clause = "pgh_context"
            annotated_context_columns_clause = "".join(
                [
                    f"(pgh_context->>'{attr}')::"
                    f"{field.rel_db_type(self.connection)} AS {field.column},\n"
                    for field, attr in proxy_fields
                ]
            )

            if hasattr(event_model, "pgh_context_id"):
                context_id_column_clause = "pgh_context_id"
            else:
                context_id_column_clause = "NULL::UUID AS pgh_context_id"
        else:
            raise AssertionError

        return (
            final_context_columns_clause,
            context_column_clause,
            context_id_column_clause,
            context_join_clause,
            annotated_context_columns_clause,
        )

    def _get_where_clause(self, event_model):
        if self.references:
            rows = self.references
            cols = [
                field.column
                for field in event_model._meta.fields
                if utils.related_model(field) == self.references_model
            ]
        elif self.tracks:
            rows = self.tracks
            cols = [event_model._meta.get_field("pgh_obj").column]
        else:
            return ""

        with self.connection.cursor() as cursor:
            if isinstance(rows, models.QuerySet) or len(rows) > 1:
                opt = "IN"
                # TODO: Use a subquery
                fmt = ",".join(["%s"] * len(rows))
                val = [o._meta.pk.get_db_prep_value(o.pk, self.connection) for o in rows]
                pks = cursor.mogrify(f"({fmt})", val)
            else:
                opt = "="
                val = rows[0]._meta.pk.get_db_prep_value(rows[0].pk, self.connection)
                pks = cursor.mogrify("%s", [val])

        if isinstance(pks, bytes):  # psycopg2 returns bytes
            pks = pks.decode("utf-8")

        return "WHERE " + " OR ".join(f"_event.{col} {opt} {pks}" for col in cols)

    def _get_select(self, event_model):
        where_clause = self._get_where_clause(event_model)

        (
            final_context_columns_clause,
            context_column_clause,
            context_id_column_clause,
            context_join_clause,
            annotated_context_columns_clause,
        ) = self._get_context_clauses(event_model)

        prev_data_clause = """
            LAG(row_to_json(_event))
              OVER (
                PARTITION BY _event.pgh_obj_id
                ORDER BY _event.pgh_id
              ) AS _prev_data
        """
        pgh_obj_id_column_clause = "pgh_obj_id::TEXT"
        if not hasattr(event_model, "pgh_obj_id"):
            prev_data_clause = "NULL::JSONB AS _prev_data"
            pgh_obj_id_column_clause = "NULL::TEXT AS pgh_obj_id"

        event_table = event_model._meta.db_table
        return f"""
            SELECT
              CONCAT('{event_model._meta.label}', ':', _pgh_obj_event.pgh_id) AS pgh_slug,
              _pgh_obj_event.pgh_id,
              _pgh_obj_event.pgh_created_at,
              _pgh_obj_event.pgh_label,
              {final_context_columns_clause}
              _pgh_obj_event.pgh_obj_id,
              '{event_model._meta.label}' AS pgh_model,
              '{event_model.pgh_tracked_model._meta.label}' AS pgh_obj_model,
              (
                  SELECT JSONB_OBJECT_AGG(filtered.key, filtered.value)
                  FROM
                    (
                        SELECT key, value
                        FROM JSONB_EACH(_pgh_obj_event._curr_data::JSONB)
                    ) filtered
                  WHERE filtered.key NOT LIKE 'pgh_%%'
              ) AS pgh_data,
              (
                SELECT JSONB_OBJECT_AGG(curr.key, array[prev.value, curr.value])
                FROM
                  (
                    SELECT key, value
                    FROM JSONB_EACH(_pgh_obj_event._curr_data::JSONB)
                  ) curr
                  LEFT OUTER JOIN
                  (
                    SELECT key, value
                    FROM JSONB_EACH(_pgh_obj_event._prev_data::JSONB)
                  ) prev
                  ON curr.key = prev.key
                WHERE curr.key NOT LIKE 'pgh_%%'
                  AND curr.value != prev.value
                  AND prev IS NOT NULL
              ) AS pgh_diff,
              _pgh_obj_event.pgh_context_id,
              _pgh_obj_event.pgh_context
            FROM (
              SELECT
                pgh_id,
                pgh_created_at,
                pgh_label,
                row_to_json(_event) AS _curr_data,
                {annotated_context_columns_clause}
                {prev_data_clause},
                {context_id_column_clause},
                {context_column_clause},
                {pgh_obj_id_column_clause}
              FROM {event_table} _event
              {context_join_clause}
              {where_clause}
              ORDER BY _event.pgh_id
            ) _pgh_obj_event
        """

    def _get_cte(self):
        """
        Returns the CTE clause for the aggregate event query
        """
        events_table = self.query.model._meta.db_table
        inner_cte = "UNION ALL ".join(
            [self._get_select(event_model) for event_model in self.across]
        )
        if not inner_cte:
            inner_cte = self._get_empty_select()

        return f"WITH {events_table} AS (\n" + inner_cte + "\n)\n"

    def as_sql(self, *args, **kwargs):
        self._validate()

        base_sql, base_params = super().as_sql(*args, **kwargs)

        # Create the CTE that will be queried and insert it into the
        # main query
        cte = self._get_cte()

        return cte + base_sql, base_params


class EventsQuery(Query):
    """A query over an aggregate event CTE"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.references = []
        self.tracks = []
        self.across = []

    def get_compiler(self, *args, **kwargs):
        compiler = super().get_compiler(*args, **kwargs)
        compiler.__class__ = EventsQueryCompiler
        return compiler

    def __chain(self, _name, klass=None, *args, **kwargs):
        clone = getattr(super(), _name)(self.__class__, *args, **kwargs)
        clone.references = self.references
        clone.tracks = self.tracks
        clone.across = self.across
        return clone

    def chain(self, klass=None):
        return self.__chain("chain", klass)


class EventsQuerySet(models.QuerySet):
    """QuerySet with support for Common Table Expressions"""

    def __init__(self, model=None, query=None, using=None, hints=None):
        # Only create an instance of a Query if this is the first invocation in
        # a query chain.
        if query is None:
            query = EventsQuery(model)

        super().__init__(model, query, using, hints)

    def across(self, *event_models):
        """Aggregates events across the provided event models"""
        qs = self._clone()
        qs.query.across = [
            apps.get_model(model) if isinstance(model, str) else model for model in event_models
        ]
        return qs

    def references(self, *objs):
        """Query any rows that reference the objs.

        If, for example, a foreign key or pgh_obj field points to the
        object, it will be aggregated.
        """
        assert len(objs) >= 1

        if isinstance(objs[0], (list, tuple, models.QuerySet)):
            assert len(objs) == 1
            objs = objs[0]

        qs = self._clone()
        qs.query.references = objs
        return qs

    def tracks(self, *objs):
        """Query any rows with pgh_obj equal to the objs."""
        assert len(objs) >= 1

        if isinstance(objs[0], (list, tuple, models.QuerySet)):
            assert len(objs) == 1
            objs = objs[0]

        qs = self._clone()
        qs.query.tracks = objs
        return qs


class NoObjectsManager(models.Manager):
    """
    Django's dumpdata and other commands will not work with Events models
    by default because of how they aggregate multiple tables based on objects.

    We use this as the default manager for aggregate events so that dumpdata
    and other management commands still work with these models
    """

    def get_queryset(self, *args, **kwargs):
        return models.QuerySet(self.model, using=self._db).none()


class Events(models.Model):
    """
    A proxy model for aggregating events together across tables and
    rendering diffs
    """

    pgh_slug = models.TextField(
        primary_key=True, help_text="The unique identifier across all event tables."
    )
    pgh_model = models.CharField(max_length=64, help_text="The event model.")
    pgh_id = models.BigIntegerField(help_text="The primary key of the event.")
    pgh_created_at = models.DateTimeField(
        auto_now_add=True, help_text="When the event was created."
    )
    pgh_label = models.TextField(help_text="The event label.")
    pgh_data = utils.JSONField(help_text="The raw data of the event.")
    pgh_diff = utils.JSONField(help_text="The diff between the previous event of the same label.")
    pgh_context_id = models.UUIDField(null=True, help_text="The context UUID.")
    pgh_context = utils.JSONField(
        null=True,
        help_text="The context associated with the event.",
    )
    pgh_obj_model = models.CharField(max_length=64, help_text="The object model.")
    pgh_obj_id = models.TextField(null=True, help_text="The primary key of the object.")

    objects = EventsQuerySet.as_manager()
    no_objects = NoObjectsManager()

    class Meta:
        managed = False
        verbose_name_plural = "events"
        # See the docs for NoObjectsManager about why this is the default
        # manager
        default_manager_name = "no_objects"

    @classmethod
    def check(cls, **kwargs):
        """
        Allow proxy models to inherit this model and define their own fields
        that are dynamically pulled from context
        """
        errors = super().check(**kwargs)
        return [error for error in errors if error.id != "models.E017"]


class MiddlewareEvents(Events):
    """
    A proxy model for aggregating events. Includes additional fields that
    are captured by the pghistory middleware
    """

    user = core.ProxyField(
        "pgh_context__user",
        models.ForeignKey(
            settings.AUTH_USER_MODEL,
            on_delete=models.DO_NOTHING,
            null=True,
            help_text="The user associated with the event.",
        ),
    )
    url = core.ProxyField(
        "pgh_context__url",
        models.TextField(null=True, help_text="The url associated with the event."),
    )

    class Meta:
        proxy = True
        verbose_name_plural = "middleware events"



import enum


class Default(enum.Enum):
    token = 0


class Unset(enum.Enum):
    token = 0


DEFAULT = Default.token
"""
For setting a configuration value back to its default value
"""

UNSET = Unset.token



import django
import pgtrigger

from pghistory.config import (
    ContextForeignKey,
    ContextJSONField,
    ContextUUIDField,
    Field,
    ForeignKey,
    ObjForeignKey,
    RelatedField,
)
from pghistory.constants import DEFAULT
from pghistory.core import (
    DeleteEvent,
    InsertEvent,
    ManualEvent,
    ProxyField,
    RowEvent,
    Tracker,
    UpdateEvent,
    create_event,
    create_event_model,
    track,
)
from pghistory.runtime import context
from pghistory.version import __version__

Condition = pgtrigger.Condition
"""For specifying free-form SQL in the condition of a trigger."""

AnyChange = pgtrigger.AnyChange
"""If any supplied fields change, trigger the event.

Args:
    *fields (str): If any supplied fields change, trigger the event.
        If no fields are supplied, defaults to all tracked fields.
    exclude (List[str]): Fields to exclude.
    exclude_auto (bool): Exclude all `auto_now` and `auto_now_add` fields automatically.
"""

AnyDontChange = pgtrigger.AnyDontChange
"""If any supplied fields don't change, trigger the event.

Args:
    *fields (str): If any supplied fields don't change, trigger the event.
        If no fields are supplied, defaults to all tracked fields.
    exclude (List[str]): Fields to exclude.
    exclude_auto (bool): Exclude all `auto_now` and `auto_now_add` fields automatically.
"""

AllChange = pgtrigger.AllChange
"""If all supplied fields change, trigger the event.

Args:
    *fields (str): If all supplied fields change, trigger the event.
        If no fields are supplied, defaults to all tracked fields.
    exclude (List[str]): Fields to exclude.
    exclude_auto (bool): Exclude all `auto_now` and `auto_now_add` fields automatically.
"""

AllDontChange = pgtrigger.AllDontChange
"""If all supplied fields don't change, trigger the event.

Args:
    *fields (str): If all supplied fields don't change, trigger the event.
        If no fields are supplied, defaults to all tracked fields.
    exclude (List[str]): Fields to exclude.
    exclude_auto (bool): Exclude all `auto_now` and `auto_now_add` fields automatically.
"""

F = pgtrigger.F
"""
Similar to Django's `F` object, allows referencing the old and new
rows in a trigger condition.
"""

Q = pgtrigger.Q
"""
Similar to Django's `Q` object, allows building filter clauses based on
the old and new rows in a trigger condition.
"""

Delete = pgtrigger.Delete
"""
For specifying `DELETE` as the trigger operation.
"""

Insert = pgtrigger.Insert
"""
For specifying `INSERT` as the trigger operation.
"""

Update = pgtrigger.Update
"""
For specifying `UPDATE` as the trigger operation.
"""

New = "NEW"
"""
For storing the trigger's "NEW" row in a [pghistory.RowEvent][]
"""

Old = "OLD"
"""
For storing the trigger's "OLD" row in a [pghistory.RowEvent][]
"""

__all__ = [
    "AnyChange",
    "AllChange",
    "AnyDontChange",
    "AllDontChange",
    "Condition",
    "context",
    "ContextForeignKey",
    "ContextJSONField",
    "ContextUUIDField",
    "create_event",
    "create_event_model",
    "DEFAULT",
    "Delete",
    "DeleteEvent",
    "F",
    "Field",
    "ForeignKey",
    "Insert",
    "InsertEvent",
    "ManualEvent",
    "New",
    "ObjForeignKey",
    "Old",
    "ProxyField",
    "Q",
    "RelatedField",
    "RowEvent",
    "track",
    "Tracker",
    "Update",
    "UpdateEvent",
    "__version__",
]

if django.VERSION < (3, 2):  # pragma: no cover
    default_app_config = "pghistory.apps.PGHistoryConfig"

del django
del pgtrigger



import django

from pghistory.admin.core import (
    EventModelAdmin,
    EventsAdmin,
)

__all__ = [
    "EventModelAdmin",
    "EventsAdmin",
]

if django.VERSION < (3, 2):  # pragma: no cover
    default_app_config = "pghistory.admin.apps.PGHistoryAdminConfig"

del django



from typing import Any, Dict

from django.core.handlers.asgi import ASGIRequest as DjangoASGIRequest
from django.core.handlers.wsgi import WSGIRequest as DjangoWSGIRequest
from django.db import connection

import pghistory
from pghistory import config


class DjangoRequest:
    """
    Although Django's auth middleware sets the user in middleware,
    apps like django-rest-framework set the user in the view layer.
    This creates issues for pghistory tracking since the context needs
    to be set before DB operations happen.

    This special WSGIRequest updates pghistory context when
    the request.user attribute is updated.
    """

    def __setattr__(self, attr, value):
        if attr == "user":
            user = (
                value._meta.pk.get_db_prep_value(value.pk, connection)
                if value and hasattr(value, "_meta")
                else None
            )
            pghistory.context(user=user)

        return super().__setattr__(attr, value)


class WSGIRequest(DjangoRequest, DjangoWSGIRequest):
    pass


class ASGIRequest(DjangoRequest, DjangoASGIRequest):
    pass


class HistoryMiddleware:
    """
    Annotates the user/url in the pghistory context.

    Add more context by inheriting the middleware and overriding the `get_context` method.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def get_context(self, request) -> Dict[str, Any]:
        user = (
            request.user._meta.pk.get_db_prep_value(request.user.pk, connection)
            if hasattr(request, "user") and hasattr(request.user, "_meta")
            else None
        )
        return {"user": user, "url": request.path}

    def __call__(self, request):
        if request.method in config.middleware_methods():
            with pghistory.context(**self.get_context(request)):
                if isinstance(request, DjangoWSGIRequest):  # pragma: no branch
                    request.__class__ = WSGIRequest
                elif isinstance(request, DjangoASGIRequest):  # pragma: no cover
                    request.__class__ = ASGIRequest

                return self.get_response(request)
        else:
            return self.get_response(request)



import re

import pgtrigger
from django.db import models

from pghistory import config, utils


def _get_pgh_obj_pk_col(history_model):
    """
    Returns the column name of the PK field tracked by the history model
    """
    return history_model._meta.get_field("pgh_obj").related_model._meta.pk.column


def _fmt_trigger_name(label):
    """Given a history event label, generate a trigger name"""
    if label:
        return re.sub("[^0-9a-zA-Z]+", "_", label).lower()
    else:  # pragma: no cover
        return None


class Event(pgtrigger.Trigger):
    """
    Events a model with a label when a condition happens
    """

    label = None
    row = "NEW"
    event_model = None
    when = pgtrigger.After

    def __init__(
        self,
        *,
        name=None,
        operation=None,
        condition=None,
        label=None,
        event_model=None,
        when=None,
        row=None,
        snapshot=None,
    ):
        # Note - "snapshot" is the old field, renamed to "row". We avoid removing it entirely
        # since old migrations still may reference this trigger
        row = row or snapshot

        self.label = label or self.label
        if not self.label:  # pragma: no cover
            raise ValueError('Must provide "label"')

        self.name = name or self.name or self.label
        if not self.name:  # pragma: no cover
            raise ValueError('Must provide "name"')

        self.name = _fmt_trigger_name(self.name)

        self.event_model = event_model or self.event_model
        if not self.event_model:  # pragma: no cover
            raise ValueError('Must provide "event_model"')

        self.row = row or self.row
        if not self.row:  # pragma: no cover
            raise ValueError('Must provide "row"')

        super().__init__(operation=operation, condition=condition, when=when)

    def get_func(self, model):
        tracked_model_fields = {f.name for f in self.event_model.pgh_tracked_model._meta.fields}
        fields = {
            f.column: f'{self.row}."{f.column}"'
            for f in self.event_model._meta.fields
            if not isinstance(f, models.AutoField)
            and f.name in tracked_model_fields
            and f.concrete
        }
        fields["pgh_created_at"] = config.created_at_function()
        fields["pgh_label"] = f"'{self.label}'"

        if hasattr(self.event_model, "pgh_obj"):
            fields["pgh_obj_id"] = f'{self.row}."{_get_pgh_obj_pk_col(self.event_model)}"'

        if hasattr(self.event_model, "pgh_context"):
            if isinstance(self.event_model._meta.get_field("pgh_context"), models.ForeignKey):
                fields["pgh_context_id"] = "_pgh_attach_context()"
            elif isinstance(self.event_model._meta.get_field("pgh_context"), utils.JSONField):
                fields["pgh_context"] = (
                    "COALESCE(NULLIF(CURRENT_SETTING('pghistory.context_metadata', TRUE), ''),"
                    " NULL)::JSONB"
                )
            else:
                raise AssertionError

        if hasattr(self.event_model, "pgh_context_id") and isinstance(
            self.event_model._meta.get_field("pgh_context_id"), models.UUIDField
        ):
            fields["pgh_context_id"] = (
                "COALESCE(NULLIF(CURRENT_SETTING('pghistory.context_id', TRUE), ''), NULL)::UUID"
            )

        fields = {key: fields[key] for key in sorted(fields)}

        cols = ", ".join(f'"{col}"' for col in fields)
        vals = ", ".join(val for val in fields.values())
        sql = f"""
            INSERT INTO "{self.event_model._meta.db_table}"
                ({cols}) VALUES ({vals});
            RETURN NULL;
        """
        return " ".join(line.strip() for line in sql.split("\n") if line.strip()).strip()



import django.apps
from django.db.models.signals import class_prepared, post_migrate

from pghistory import config


def pgh_setup(sender, **kwargs):
    if hasattr(sender, "pghistory_setup"):
        sender.pghistory_setup()


def install_on_migrate(using, **kwargs):  # pragma: no cover
    if config.install_context_func_on_migrate():
        Context = django.apps.apps.get_model("pghistory", "Context")
        Context.install_pgh_attach_context_func(using=using)


class PGHistoryConfig(django.apps.AppConfig):
    name = "pghistory"

    def __init__(self, *args, **kwargs):
        """
        Install pgtriggers for Change detection models whenever
        the change detection model class is ready. We have to do this
        in __init__ instead of ready() here since the class_prepared
        signal is emitted before models are ready
        """
        class_prepared.connect(pgh_setup)
        super().__init__(*args, **kwargs)

    def ready(self):
        # Register custom checks
        from pghistory import checks  # noqa

        post_migrate.connect(install_on_migrate, sender=self)



# Generated by Django 4.2.6 on 2023-10-14 20:31

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("pghistory", "0005_events_middlewareevents"),
    ]

    operations = [
        migrations.DeleteModel(
            name="AggregateEvent",
        ),
    ]



import collections

from django import template, urls
from django.apps import apps
from django.contrib import admin

from pghistory import config, models
from pghistory.admin import EventModelAdmin

register = template.Library()


@register.filter
def events_are_tracked(model):
    model = apps.get_model(model)
    return any(
        model._meta.concrete_model == m.pgh_tracked_model._meta.concrete_model
        for m in apps.get_models()
        if issubclass(m, models.Event) and m.pgh_tracked_model
    )


@register.simple_tag()
def events_admin_url():
    """Retrieves the admin url of the events"""
    model = config.admin_queryset().model

    return urls.reverse(f"admin:{model._meta.app_label}_{model._meta.model_name}_changelist")


@register.filter
def event_admins(model):
    """Retrieves the admins of all event models associated with the primary model"""
    model = apps.get_model(model)._meta.concrete_model
    tracked_models = collections.defaultdict(set)

    for m in apps.get_models():
        if issubclass(m, models.Event) and m.pgh_tracked_model and hasattr(m, "pgh_obj"):
            tracked_models[m.pgh_tracked_model._meta.concrete_model].add(m._meta.concrete_model)

    admin_registry = {
        m._meta.concrete_model: admin
        for m, admin in admin.site._registry.items()
        if isinstance(admin, EventModelAdmin)
    }
    return [admin_registry[m].model._meta for m in tracked_models[model] if m in admin_registry]



# Generated by Django 4.1 on 2022-10-06 00:05
# flake8: noqa

from django.db import migrations, models
import django.db.models.manager
import pghistory.models


class Migration(migrations.Migration):
    dependencies = [
        ("pghistory", "0004_auto_20220906_1625"),
    ]

    operations = [
        migrations.CreateModel(
            name="Events",
            fields=[
                (
                    "pgh_slug",
                    models.TextField(
                        help_text="The unique identifier across all event tables.",
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("pgh_model", models.CharField(help_text="The event model.", max_length=64)),
                ("pgh_id", models.BigIntegerField(help_text="The primary key of the event.")),
                (
                    "pgh_created_at",
                    models.DateTimeField(
                        auto_now_add=True, help_text="When the event was created."
                    ),
                ),
                ("pgh_label", models.TextField(help_text="The event label.")),
                (
                    "pgh_data",
                    pghistory.models.PGHistoryJSONField(help_text="The raw data of the event."),
                ),
                (
                    "pgh_diff",
                    pghistory.models.PGHistoryJSONField(
                        help_text="The diff between the previous event of the same label."
                    ),
                ),
                ("pgh_context_id", models.UUIDField(help_text="The context UUID.", null=True)),
                (
                    "pgh_context",
                    pghistory.models.PGHistoryJSONField(
                        help_text="The context associated with the event.", null=True
                    ),
                ),
                ("pgh_obj_model", models.CharField(help_text="The object model.", max_length=64)),
                (
                    "pgh_obj_id",
                    models.TextField(help_text="The primary key of the object.", null=True),
                ),
            ],
            options={
                "verbose_name_plural": "events",
                "managed": False,
                "default_manager_name": "no_objects",
            },
            managers=[
                ("no_objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name="MiddlewareEvents",
            fields=[],
            options={
                "verbose_name_plural": "middleware events",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("pghistory.events",),
            managers=[
                ("no_objects", django.db.models.manager.Manager()),
            ],
        ),
    ]



"""Core functionality and interface of pghistory"""

import copy
import re
import sys
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

import pgtrigger
import pgtrigger.core
from django.apps import apps
from django.db import connections, models
from django.db.models import sql
from django.db.models.fields.related import RelatedField
from django.db.models.sql import compiler
from django.utils.module_loading import import_string
from django.utils.text import slugify

from pghistory import config, constants, runtime, trigger, utils

if TYPE_CHECKING:
    from pghistory import ContextForeignKey, ContextJSONField, ContextUUIDField, ObjForeignKey

if utils.psycopg_maj_version == 2:
    from psycopg2.extensions import AsIs as Literal
elif utils.psycopg_maj_version == 3:
    import psycopg.adapt

    class Literal:
        def __init__(self, val):
            self.val = val

    class LiteralDumper(psycopg.adapt.Dumper):
        def dump(self, obj):
            return obj.val.encode("utf-8")

        def quote(self, obj):
            return self.dump(obj)

else:
    raise AssertionError


_registered_trackers = {}


class Tracker:
    """For tracking an event when a condition happens on a model."""

    label: Optional[str] = None

    def __init__(self, label=None):
        self.label = label or self.label

        if not self.label:  # pragma: no cover
            raise ValueError("Must supply label attribute to event")

    def setup(self, event_model):
        """Set up the tracker for the event model"""
        pass

    def pghistory_setup(self, event_model):
        """Registers the tracker for the event model and calls user-defined setup"""
        tracked_model = event_model.pgh_tracked_model

        if _registered_trackers.get((tracked_model, self.label), event_model) != event_model:
            raise ValueError(
                f'Tracker with label "{self.label}" already exists for a different'
                f' event model of "{tracked_model._meta.label}". Supply a'
                " different label as the first argument to the tracker."
            )

        _registered_trackers[(tracked_model, self.label)] = event_model

        self.setup(event_model)


class ManualEvent(Tracker):
    """For manually tracking an event."""


class RowEvent(Tracker):
    """For tracking an event automatically based on row-level changes."""

    condition: Union[Optional[pgtrigger.Condition], constants.Unset] = constants.UNSET
    operation: Optional[pgtrigger.Operation] = None
    row: Optional[str] = None
    trigger_name: Optional[str] = None

    def __init__(
        self,
        label: Optional[str] = None,
        *,
        condition: Union[Optional[pgtrigger.Condition], constants.Unset] = constants.UNSET,
        operation: Optional[pgtrigger.Operation] = None,
        row: Optional[str] = None,
        trigger_name: Optional[str] = None,
    ):
        super().__init__(label=label)

        self.condition = condition or self.condition
        self.operation = operation or self.operation
        self.row = row or self.row
        self.trigger_name = trigger_name or self.trigger_name or f"{self.label}_{self.operation}"

        if self.condition is constants.UNSET:
            self.condition = pgtrigger.AnyChange() if self.operation == pgtrigger.Update else None

        if self.row not in ("OLD", "NEW"):  # pragma: no cover
            raise ValueError("Row must be specified as pghistory.Old or pghistory.New")
        elif self.operation == pgtrigger.Insert and self.row == "OLD":  # pragma: no cover
            raise ValueError('There is no "OLD" row on insert events')
        elif self.operation == pgtrigger.Delete and self.row == "NEW":  # pragma: no cover
            raise ValueError('There is no "NEW" row on delete events')

        if not self.operation:  # pragma: no cover
            raise ValueError("Must provide operation to RowEvent")

    def add_event_trigger(self, event_model):
        pgtrigger.register(
            trigger.Event(
                event_model=event_model,
                label=self.label,
                name=self.trigger_name,
                row=self.row,
                operation=self.operation,
                condition=self.condition,
            )
        )(event_model.pgh_tracked_model)

    def setup(self, event_model):
        # If any _Change condition is used, modify the default fields to the tracked fields
        if isinstance(self.condition, pgtrigger.core._Change) and not self.condition.fields:
            self.condition = copy.deepcopy(self.condition)
            model_fields = {f.name for f in event_model.pgh_tracked_model._meta.fields}
            self.condition.fields = [
                field.name for field in event_model._meta.fields if field.name in model_fields
            ]

        self.add_event_trigger(event_model)


class InsertEvent(RowEvent):
    """Creates events based on inserts to a model

    The default label used is "insert".
    """

    label: Optional[str] = "insert"
    row: Optional[str] = "NEW"
    operation: Optional[pgtrigger.Operation] = pgtrigger.Insert


class UpdateEvent(RowEvent):
    """Creates events based on updates to a model.

    By default,

    - The label used is "update".
    - Attributes from the "new" row of the update are stored.
    - It only fires when fields of the event model are changed.

    All of this behavior can be overridden by supplying a label,
    a condition, or the row to snapshot.
    """

    label: Optional[str] = "update"
    row: Optional[str] = "NEW"
    operation: Optional[pgtrigger.Operation] = pgtrigger.Update


class DeleteEvent(RowEvent):
    """Creates events based on deletes to a model

    The default label used is "delete".
    """

    label: Optional[str] = "delete"
    row: Optional[str] = "OLD"
    operation: Optional[pgtrigger.Operation] = pgtrigger.Delete


def _pascalcase(string):
    """Convert string into pascal case."""

    string = re.sub(r"^[\-_\.]", "", str(string))
    if not string:  # pragma: no branch
        return string

    return string[0].upper() + re.sub(
        r"[\-_\.\s]([a-z])",
        lambda matched: matched.group(1).upper(),
        string[1:],
    )


def _generate_event_model_name(base_model, tracked_model, fields):
    """Generates a default history model name"""
    name = tracked_model._meta.object_name
    if fields:
        name += "_" + "_".join(fields)

    name += f"_{base_model._meta.object_name.lower()}"
    return _pascalcase(name)


def _get_field_construction(field):
    _, _, args, kwargs = field.deconstruct()

    if isinstance(field, models.ForeignKey):
        default = config.foreign_key_field()
    elif isinstance(field, RelatedField):  # pragma: no cover
        default = config.related_field()
    else:
        default = config.field()

    kwargs.update(default.kwargs)

    cls = field.__class__
    if isinstance(field, models.OneToOneField):
        cls = models.ForeignKey
    elif isinstance(field, models.FileField):
        kwargs.pop("primary_key", None)

    for field_class, exclude_kwargs in config.exclude_field_kwargs().items():
        if isinstance(field, field_class):
            for exclude_kwarg in exclude_kwargs:
                kwargs.pop(exclude_kwarg, None)

    return cls, args, kwargs


def _generate_history_field(tracked_model, field):
    """
    When generating a history model from a tracked model, ensure the fields
    are set up properly so that related names and other information
    from the tracked model do not cause errors.
    """
    field = tracked_model._meta.get_field(field)

    if isinstance(field, models.BigAutoField):
        return models.BigIntegerField()
    elif isinstance(field, models.AutoField):
        return models.IntegerField()
    elif not field.concrete:  # pragma: no cover
        # Django doesn't have any non-concrete fields that appear
        # in ._meta.fields, but packages like django-prices have
        # non-concrete fields
        return field

    # The "swappable" field causes issues during deconstruct()
    # since it tries to load models. Patch it and set it back to the original
    # value later
    field = copy.deepcopy(field)
    swappable = getattr(field, "swappable", constants.UNSET)
    field.swappable = False
    cls, args, kwargs = _get_field_construction(field)
    field = cls(*args, **kwargs)

    if swappable is not constants.UNSET:
        field.swappable = swappable

    return field


def _generate_related_name(base_model, fields):
    """
    Generates a related name to the tracking model based on the base
    model and traked fields
    """
    related_name = slugify(base_model._meta.verbose_name_plural).replace("-", "_")
    return "_".join(fields) + f"_{related_name}" if fields else related_name


def _validate_event_model_path(*, app_label, model_name, abstract):
    if app_label not in apps.app_configs:
        raise ValueError(f'App label "{app_label}" is invalid')

    app = apps.app_configs[app_label]
    models_module = app.module.__name__ + ".models"
    if not abstract and hasattr(sys.modules[models_module], model_name):
        raise ValueError(
            f"App {app_label} already has {model_name} model. You must"
            " explicitly declare an unused model name for the pghistory model."
        )
    elif models_module.startswith("django."):
        raise ValueError(
            "A history model cannot be generated under third party app"
            f' "{app_label}". You must explicitly pass an app label'
            " when configuring tracking."
        )


def _get_obj_field(*, obj_field, tracked_model, base_model, fields):
    if obj_field is None:  # pragma: no cover
        return None
    elif obj_field is constants.UNSET:
        obj_field = config.obj_field()
        if obj_field._kwargs.get("related_name", constants.DEFAULT) == constants.DEFAULT:
            obj_field._kwargs["related_name"] = _generate_related_name(base_model, fields)

    if isinstance(obj_field, config.ObjForeignKey):
        return models.ForeignKey(tracked_model, **obj_field.kwargs)
    else:  # pragma: no cover
        raise TypeError("obj_field must be of type pghistory.ObjForeignKey.")


def _get_context_field(context_field):
    if context_field is None:  # pragma: no cover
        return None
    elif context_field is constants.UNSET:
        context_field = config.context_field()

    if isinstance(context_field, config.ContextForeignKey):
        return models.ForeignKey("pghistory.Context", **context_field.kwargs)
    elif isinstance(context_field, config.ContextJSONField):
        return utils.JSONField(**context_field.kwargs)
    else:  # pragma: no cover
        raise TypeError(
            "context_field must be of type pghistory.ContextForeignKey"
            " or pghistory.ContextJSONField."
        )


def _get_context_id_field(context_id_field):
    if context_id_field is None:
        return None
    elif context_id_field is constants.UNSET:  # pragma: no branch
        context_id_field = config.context_id_field()

    if isinstance(context_id_field, config.ContextUUIDField):
        return models.UUIDField(**context_id_field.kwargs)
    else:  # pragma: no cover
        raise TypeError("context_id_field must be of type pghistory.ContextUUIDField.")


def _get_append_only(append_only):
    return config.append_only() if append_only is constants.UNSET else append_only


def create_event_model(
    tracked_model: Type[models.Model],
    *trackers: Tracker,
    fields: Union[List[str], None] = None,
    exclude: Union[List[str], None] = None,
    obj_field: Union["ObjForeignKey", constants.Unset] = constants.UNSET,
    context_field: Union[
        "ContextForeignKey", "ContextJSONField", constants.Unset
    ] = constants.UNSET,
    context_id_field: Union["ContextUUIDField", constants.Unset] = constants.UNSET,
    append_only: Union[bool, constants.Unset] = constants.UNSET,
    model_name: Union[str, None] = None,
    app_label: Union[str, None] = None,
    base_model: Optional[Type[models.Model]] = None,
    attrs: Optional[Dict[str, Any]] = None,
    meta: Optional[Dict[str, Any]] = None,
    abstract: bool = True,
) -> Type[models.Model]:
    """
    Create an event model.

    Instead of using [pghistory.track][], which dynamically generates an event
    model, one can instead construct a event model themselves, which
    will also set up event tracking for the tracked model.

    Args:
        tracked_model: The model that is being tracked.
        *trackers: The event trackers. When using any tracker that inherits
            [pghistory.RowEvent][], such as [pghistory.InsertEvent][], a
            Postgres trigger will be installed that automatically stores the event
            into the generated event model. Trackers that do not inherit
            [pghistory.RowEvent][] must be manually created. If no events are
            supplied, defaults to `pghistory.InsertEvent` and `pghistory.UpdateEvent`.
        fields: The list of fields to store when the event takes place. When
            no fields are provided, all fields are stored. Note that storing
            the OLD or NEW row is configured by the `row` attribute of the `RowEvent` object.
            Manual events must specify these fields during manual creation.
        exclude: Instead of providing a list of fields to snapshot, a user can
            instead provide a list of fields to not snapshot.
        obj_field: The foreign key field configuration that references the tracked object.
            Defaults to an unconstrained non-nullable foreign key. Use `None` to create a
            event model with no reference to the tracked object.
        context_field: The context field configuration. Defaults to a nullable
            unconstrained foreign key. Use `None` to avoid attaching historical context altogether.
        context_id_field: The context ID field configuration when using a ContextJSONField
            for the context_field. When using a denormalized context field, the ID
            field is used to track the UUID of the context. Use `None` to avoid using this
            field for denormalized context.
        append_only: True if the event model is protected against updates and deletes.
        model_name: Use a custom model name when the event model is generated. Otherwise
            a default name based on the tracked model and fields will be created.
        app_label: The app_label for the generated event model. Defaults to the app_label
            of the tracked model. Note, when tracking a Django model (User) or a model
            of a third-party app, one must manually specify the app_label of an internal
            app to use so that migrations work properly.
        base_model: The base model for the event model. Must inherit pghistory.models.Event.
        attrs: Additional attributes to add to the event model
        meta: Additional attributes to add to the Meta class of the event model.
        abstract: `True` if the generated model should be an abstract model.

    Returns:
        The event model class.

    Example:
        Create a custom event model:

            class MyEventModel(create_event_model(
                TrackedModel,
                pghistory.InsertEvent(),
            )):
                # Add custom indices or change default field declarations...
    """  # noqa
    if not trackers:
        trackers = config.default_trackers() or (InsertEvent(), UpdateEvent())

    event_model = import_string("pghistory.models.Event")
    base_model = base_model or config.base_model()
    assert issubclass(base_model, event_model)

    obj_field = _get_obj_field(
        obj_field=obj_field,
        tracked_model=tracked_model,
        base_model=base_model,
        fields=fields,
    )
    context_field = _get_context_field(context_field)
    context_id_field = _get_context_id_field(context_id_field)
    append_only = _get_append_only(append_only)

    model_name = model_name or _generate_event_model_name(base_model, tracked_model, fields)
    app_label = app_label or tracked_model._meta.app_label
    _validate_event_model_path(app_label=app_label, model_name=model_name, abstract=abstract)
    app = apps.app_configs[app_label]
    models_module = app.module.__name__ + ".models"

    attrs = attrs or {}
    attrs.update({"pgh_trackers": trackers})
    meta = meta or {}
    exclude = exclude or []
    all_fields = (tracked_model._meta.concrete_model or tracked_model)._meta.local_fields
    fields = (
        fields if fields is not None else [f.name for f in all_fields if f.name not in exclude]
    )

    if append_only:
        meta["triggers"] = [
            *meta.get("triggers", []),
            pgtrigger.Protect(name="append_only", operation=pgtrigger.Update | pgtrigger.Delete),
        ]

    class_attrs = {
        "__module__": models_module,
        "Meta": type("Meta", (), {"abstract": abstract, "app_label": app_label, **meta}),
        "pgh_tracked_model": tracked_model,
        **{field: _generate_history_field(tracked_model, field) for field in fields},
        **attrs,
    }

    if isinstance(context_field, utils.JSONField) and context_id_field:
        class_attrs["pgh_context_id"] = context_id_field

    if context_field:
        class_attrs["pgh_context"] = context_field

    if obj_field:
        class_attrs["pgh_obj"] = obj_field

    event_model = type(model_name, (base_model,), class_attrs)
    if not abstract:
        setattr(sys.modules[models_module], model_name, event_model)

    return event_model


def ProxyField(proxy: str, field: Type[models.Field]):
    """
    Proxies a JSON field from a model and adds it as a field in the queryset.

    Args:
        proxy: The value to proxy, e.g. "user__email"
        field: The field that will be used to cast the resulting value
    """
    if not isinstance(field, models.Field):  # pragma: no cover
        raise TypeError(f'"{field}" is not a Django model Field instace')

    field.pgh_proxy = proxy
    return field


def track(
    *trackers: Tracker,
    fields: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    obj_field: Union[Optional["ObjForeignKey"], constants.Unset] = constants.UNSET,
    context_field: Union[
        "ContextForeignKey", "ContextJSONField", constants.Unset
    ] = constants.UNSET,
    context_id_field: Union["ContextUUIDField", constants.Unset] = constants.UNSET,
    append_only: Union[bool, constants.Unset] = constants.UNSET,
    model_name: Optional[str] = None,
    app_label: Optional[str] = None,
    base_model: Optional[Type[models.Model]] = None,
    attrs: Optional[Dict[str, Any]] = None,
    meta: Optional[Dict[str, Any]] = None,
):
    """
    A decorator for tracking events for a model.

    When using this decorator, an event model is dynamically generated
    that snapshots the entire model or supplied fields of the model
    based on the `events` supplied. The snapshot is accompanied with
    the label that identifies the event.

    Args:
        *trackers: The event trackers. When using any tracker that inherits
            [pghistory.RowEvent][], such as [pghistory.InsertEvent][], a
            Postgres trigger will be installed that automatically stores the event
            into the generated event model. Trackers that do not inherit
            [pghistory.RowEvent][] must be manually created. If no events are
            supplied, defaults to `pghistory.InsertEvent` and `pghistory.UpdateEvent`.
        fields: The list of fields to snapshot when the event takes place. When no fields
            are provided, the entire model is snapshot when the event happens. Note that
            snapshotting of the OLD or NEW row is configured by the `snapshot`
            attribute of the `DatabaseTracker` object. Manual events must specify
            these fields during manual creation.
        exclude: Instead of providing a list of fields to snapshot, a user can instead
            provide a list of fields to not snapshot.
        obj_field: The foreign key field configuration that references the tracked object.
            Defaults to an unconstrained non-nullable foreign key. Use `None` to create a
            event model with no reference to the tracked object.
        context_field: The context field configuration. Defaults to a nullable unconstrained
            foreign key. Use `None` to avoid attaching historical context altogether.
        context_id_field: The context ID field configuration when using a ContextJSONField for
            the context_field. When using a denormalized context field, the ID field is used to
            track the UUID of the context. Use `None` to avoid using this field for denormalized
            context.
        append_only: True if the event model is protected against updates and deletes.
        model_name: Use a custom model name when the event model is generated. Otherwise a default
            name based on the tracked model and fields will be created.
        app_label: The app_label for the generated event model. Defaults to the app_label of the
            tracked model. Note, when tracking a Django model (User) or a model of a third-party
            app, one must manually specify the app_label of an internal app to
            use so that migrations work properly.
        base_model: The base model for the event model. Must inherit `pghistory.models.Event`.
        attrs: Additional attributes to add to the event model
        meta: Additional attributes to add to the Meta class of the event model.
    """

    def _model_wrapper(model_class):
        create_event_model(
            model_class,
            *trackers,
            fields=fields,
            exclude=exclude,
            obj_field=obj_field,
            context_field=context_field,
            context_id_field=context_id_field,
            append_only=append_only,
            model_name=model_name,
            app_label=app_label,
            abstract=False,
            base_model=base_model,
            attrs=attrs,
            meta=meta,
        )

        return model_class

    return _model_wrapper


class _InsertEventCompiler(compiler.SQLInsertCompiler):
    def as_sql(self, *args, **kwargs):
        ret = super().as_sql(*args, **kwargs)
        assert len(ret) == 1
        params = [
            param if field.name != "pgh_context" else Literal("_pgh_attach_context()")
            for field, param in zip(self.query.fields, ret[0][1])
        ]

        return [(ret[0][0], params)]


def create_event(obj: models.Model, *, label: str, using: str = "default") -> models.Model:
    """Manually create a event for an object.

    Events are automatically linked with any context being tracked
    via [pghistory.context][].

    Args:
        obj: An instance of a model.
        label: The event label.
        using: The database

    Raises:
        ValueError: If the event label has not been registered for the model.

    Returns:
        The created event model object
    """
    # Verify that the provided label is tracked
    if (obj.__class__, label) not in _registered_trackers:
        raise ValueError(
            f'"{label}" is not a registered tracker label for model {obj._meta.object_name}.'
        )

    event_model = _registered_trackers[(obj.__class__, label)]
    tracked_model_fields = {field.attname for field in obj._meta.fields}
    event_model_kwargs = {
        "pgh_label": label,
        **{
            field.attname: getattr(obj, field.attname)
            for field in event_model._meta.fields
            if not field.name.startswith("pgh_") and field.attname in tracked_model_fields
        },
    }
    if hasattr(event_model, "pgh_obj"):
        event_model_kwargs["pgh_obj"] = obj

    if hasattr(event_model, "pgh_context") and isinstance(
        event_model.pgh_context.field, utils.JSONField
    ):
        if hasattr(runtime._tracker, "value"):
            event_model_kwargs["pgh_context"] = runtime._tracker.value.metadata

            if hasattr(event_model, "pgh_context_id"):
                event_model_kwargs["pgh_context_id"] = runtime._tracker.value.id

        return event_model.objects.create(**event_model_kwargs)
    else:
        event_obj = event_model(**event_model_kwargs)

        # The event model is inserted manually with a custom SQL compiler
        # that attaches the context using the _pgh_attach_context
        # stored procedure. Django does not allow one to use F()
        # objects to reference stored procedures, so we have to
        # inject it with a custom SQL compiler here.
        query = sql.InsertQuery(event_model)
        query.insert_values(
            [
                field
                for field in event_model._meta.fields
                if not isinstance(field, models.AutoField)
            ],
            [event_obj],
        )

        if utils.psycopg_maj_version == 3:
            connections[using].connection.adapters.register_dumper(Literal, LiteralDumper)

        vals = _InsertEventCompiler(query, connections[using], using=using).execute_sql(
            event_model._meta.fields
        )

        # Django >= 3.1 returns the values as a list of one element
        if isinstance(vals, list) and len(vals) == 1:  # pragma: no branch
            vals = vals[0]

        for field, val in zip(event_model._meta.fields, vals):
            setattr(event_obj, field.attname, val)

        return event_obj


def event_models(
    models: Optional[List[Type[models.Model]]] = None,
    references_model: Optional[Type[models.Model]] = None,
    tracks_model: Optional[Type[models.Model]] = None,
    include_missing_pgh_obj: bool = False,
) -> List[Type[models.Model]]:
    """
    Retrieve and filter all events models.

    Args:
        models: The starting list of event models.
        references_model: Filter by event models that reference this model.
        tracks_model: Filter by models that directly track this model and have pgh_obj fields
        include_missing_pgh_obj: Return tracked models even if the pgh_obj field is not
            available.

    Returns:
        The list of event models
    """
    from pghistory.models import Event  # noqa

    models = models or [
        model
        for model in apps.get_models()
        if issubclass(model, Event)
        and not model._meta.abstract
        and not model._meta.proxy
        and model._meta.managed
    ]

    if references_model:
        if references_model._meta.proxy:  # pragma: no cover
            references_model = references_model._meta.concrete_model

        models = [
            model
            for model in models
            if any(utils.related_model(field) == references_model for field in model._meta.fields)
        ]

    if tracks_model:
        if tracks_model._meta.proxy:  # pragma: no cover
            tracks_model = tracks_model._meta.concrete_model

        if not include_missing_pgh_obj:
            models = [
                model
                for model in models
                if "pgh_obj" in (f.name for f in model._meta.fields)
                and utils.related_model(model._meta.get_field("pgh_obj")) == tracks_model
            ]
        else:
            models = [
                model
                for model in models
                if model.pgh_tracked_model._meta.concrete_model == tracks_model
            ]

    return models



import django
from django.core.exceptions import ImproperlyConfigured
from django.utils.version import get_version_tuple

# Django>=3.1 changes the location of JSONField
if django.VERSION >= (3, 1):
    from django.db.models import JSONField as DjangoJSONField
else:  # pragma: no cover
    from django.contrib.postgres.fields import JSONField as DjangoJSONField


def _psycopg_version():
    try:
        import psycopg as Database
    except ImportError:
        import psycopg2 as Database
    except Exception as exc:  # pragma: no cover
        raise ImproperlyConfigured("Error loading psycopg2 or psycopg module") from exc

    version_tuple = get_version_tuple(Database.__version__.split(" ", 1)[0])

    if version_tuple[0] not in (2, 3):  # pragma: no cover
        raise ImproperlyConfigured(f"Pysocpg version {version_tuple[0]} not supported")

    return version_tuple


psycopg_version = _psycopg_version()
psycopg_maj_version = psycopg_version[0]


class JSONField(DjangoJSONField):
    """
    Creates a consistent import path for JSONField regardless of Django
    version.
    """


def related_model(field):
    """Return the concrete model a field references"""
    if hasattr(field, "related_model") and field.related_model:
        return field.related_model._meta.concrete_model



import django.apps
from django.contrib import admin

from pghistory import config


class PGHistoryAdminConfig(django.apps.AppConfig):
    name = "pghistory.admin"
    label = "pghistory_admin"

    def ready(self):
        admin.site.register(config.admin_queryset().model, config.admin_class())
